#!/usr/bin/perl
#
# $Id$
#
# (C) ITEG IT-Engineers GmbH 2016
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

use strict;

#use HTTP::Request;
#use HTTP::Request::Common;
#use List::MoreUtils qw/ uniq /;
#use LWP::UserAgent;

require "/usr/lib/pba/pba-functions.pl";
require "/usr/lib/pba/build-functions.pl";


sub initRpmBuildFunctions() {

  # nothing rpm specific yet

}


sub checkRpmBuildConfig() {

  # nothing rpm specific yet

  $main::cacheDir = "/var/cache/pba-builder/rpm";
  $main::libDir = "/var/lib/pba-builder/rpm";

  # signature options
  $main::createSignedRpms = 0;
  if ($main::gpgKeyId eq "") {
    ($main::verbosity > 0) && print "No GPG key configured, not signing rpm packages.\n";
  } else {
    if ($main::opts{"u"}) {
      ($main::verbosity > 0) && print "Creating unsigned rpm packages despite GPG key, due to -u.\n";
    } else {
      ($main::verbosity > 0) && print "Using GPG key with ID ".$main::gpgKeyId." to sign rpm packages.\n";
      $main::createSignedRpms = 1;
    }
  }
  
}


sub scanForIncomingSrpmFiles() {
  #%main::srpmFilesToBuildByTimestamp = ();
  $main::srpmFileToBuild = "";
  my $oldestTimestamp = time() + 1;
  my @dirsToScan = split / +/, $main::incomingDirs;
  foreach my $dirToScan (@dirsToScan) {
    ($main::verbosity > 2) && print "Scanning incoming directory ".$dirToScan."\n";
    chdir $dirToScan || cleanupAndDie("FATAL: Cannot chdir to ".$dirToScan."\n");
    my @files = glob("*.src.rpm");
    foreach my $file (@files) {
      next if (! -f $file); # skip dirs
      ($main::verbosity > 2) && print "Found source package: ".$file."\n";
      my $fileTimestamp = (stat $file)[9];
      if ($fileTimestamp lt $oldestTimestamp) { 
        $main::srpmFileToBuild = $dirToScan."/".$file;
        $oldestTimestamp = $fileTimestamp;
      }
    }
  }
  ( ($main::verbosity > 2) and ($main::srpmFileToBuild ne "") ) && print "Oldest found source package is: ".$main::srpmFileToBuild."\n";
}



sub extractRpmBuilder( $ $ $ ) {
  # extracts a rpmbuild tarball to $cacheDir/$$/$targetDistro
  # $1 = target distribution
  # $2 = logfile
  # $3 = architecture

  my $targetDistribution = shift;
  ($targetDistribution eq "") && cleanupAndDie("FATAL: extractRpmBuilder called with empty targetdistribution.\n");
  my $rpmBuilderLog = shift;
  ($rpmBuilderLog eq "") && cleanupAndDie("FATAL: extractRpmBuilder called with empty logfile.\n");
  my $architecture = shift;
  #($architecture eq "") && cleanupAndDie("FATAL: extractRpmBuilder called with empty architecture.\n");

  my $workdir = $main::cacheDir."/".$$;
  if (-d $workdir."/".$targetDistribution) {
    cleanupAndDie("FATAL: Workdir ".$workdir."/".$targetDistribution." already exists.");
  }
  checkDirectory($workdir);

  ## reactivate:
  push @main::dirsToCleanup, $workdir;

  chdir $workdir || cleanupAndDie("FATAL: Cannot chdir to ".$workdir);

  my $rpmBuilderTee = " 2>&1 |tee ".$rpmBuilderLog;

  my $tarball = $main::libDir."/".$targetDistribution.".tar.gz";

  my $tarCommand = "tar -xzf ".$tarball.$rpmBuilderTee;
  ($main::verbosity > 1) && print "Executing this tar command while in ".$workdir.":\n".$tarCommand."\n";
  executeOrDie($tarCommand);
  
  ($main::verbosity > 1) && print "Successfully extracted ".$tarball." to ".$workdir."/".$targetDistribution."\n";

}



sub retarRpmBuilder( $ $ $ ) {
  # re-tars a rpmbuild tarball from an eventual updated $cacheDir/$$/$targetDistro
  # $1 = target distribution
  # $2 = logfile
  # $3 = architecture

  my $targetDistribution = shift;
  ($targetDistribution eq "") && cleanupAndDie("FATAL: retarRpmBuilder called with empty targetdistribution.\n");
  my $rpmBuilderLog = shift;
  ($rpmBuilderLog eq "") && cleanupAndDie("FATAL: retarRpmBuilder called with empty logfile.\n");
  my $architecture = shift;
  #($architecture eq "") && cleanupAndDie("FATAL: retarRpmBuilder called with empty architecture.\n");

  my $workdir = $main::cacheDir."/".$$;
  chdir $workdir || cleanupAndDie("FATAL: Cannot chdir to ".$workdir);

  my $rpmBuilderTee = " 2>&1 |tee -a ".$rpmBuilderLog;

  my $tarball = $main::libDir."/".$targetDistribution.".tar.gz";
  if (-f $tarball) {
    executeOrDie("/bin/mv -v ".$tarball." ".$tarball.".bak");
  }

  my $tarCommand = "tar -czSf ".$tarball." ".$targetDistribution.$rpmBuilderTee;
  ($main::verbosity > 0) && print "Executing this tar command while in ".$workdir.":\n".$tarCommand."\n";
  executeOrDie($tarCommand);

  ($main::verbosity > 1) && print "Successfully re-tar-ed ".$tarball." from ".$workdir."/".$targetDistribution."\n";
}


sub executeInRpmBuilder( $ $ $ $ ) {
  # executes a script "in" a pbabuilder
  # $1 = script text (including apropriate line 1 like #!/bin/sh) or local file, or "-" to ust execute a shell in there
  # $2 = target distribution
  # $3 = logfile
  ## $4 = architecture
  # dies if parameters are empty
  # returns ...

  my $script = shift;
  ($script eq "") && cleanupAndDie("FATAL: executeInPbuilder called with empty script string.\n");
  my $targetDistribution = shift;
  ($targetDistribution eq "") && cleanupAndDie("FATAL: executeInPbuilder called with empty targetdistribution.\n");
  my $rpmBuilderLog = shift;
  #($rpmBuilderLog eq "") && cleanupAndDie("FATAL: executeInPbuilder called with empty rpmBuilderLog parameter.\n");
  if (length($rpmBuilderLog) > 0) {
    if (-f $rpmBuilderLog) {
      executeOrDie("/bin/mv ".$rpmBuilderLog." ".$rpmBuilderLog.".bak");
    }
  }
  my $architecture = shift;
  # may be empty
  #($architecture eq "") && cleanupAndDie("FATAL: executeInPbuilder called with empty architecture.\n");

  my $rpmBuilderTee = " 2>&1 |tee -a ".$rpmBuilderLog;
  if (length($rpmBuilderLog) == 0) {
    $rpmBuilderTee = "";
  }

  my $scriptOptions = "";

  my $workdir = $main::cacheDir."/".$$;
  chdir $workdir."/".$targetDistribution || cleanupAndDie("FATAL: Cannot chdir to ".$workdir."/".$targetDistribution);
  
  my $scriptFile = "script4pba-rpmbilder.sh";
  my $scriptFileToRm = "script4pba-rpmbilder.sh";
  if ( $script eq "-" ) {
    $scriptFile = "-l";
    $scriptFileToRm = "";
    $rpmBuilderTee = "";
  } else {
    if ( ($script !~ /\n/) and ($script !~ /\//) and (-x $script) ) {
      #($main::verbosity > 2) && print "executeInPbuilder: Interpretating \$1 (".$scriptFile.") as script, not as set of commands.\n"; 
      executeOrDie("/bin/cp -a ".$script." ".$scriptFile);
    } else {
      if (-f $scriptFile) { unlink $scriptFile || cleanupAndDie("FATAL: Cannot delete obsolete ".$scriptFile."\n"); }
      printToFile( $scriptFile, $script, 0755 );
    }
  }

  chdir $workdir || cleanupAndDie("FATAL: Cannot chdir to ".$workdir);

  executeOrDie("mount --bind /dev ".$targetDistribution."/dev".$rpmBuilderTee);
  push @main::mountsToUnmount, $workdir."/".$targetDistribution."/dev";
  executeOrDie("mount --bind /proc ".$targetDistribution."/proc".$rpmBuilderTee);
  push @main::mountsToUnmount, $workdir."/".$targetDistribution."/proc";

  my $chrootCommand = "/usr/sbin/chroot ".$targetDistribution." /bin/bash ".$scriptFile.$rpmBuilderTee;
  #my $hostArchitecture = myArchitecture();
  #if (($hostArchitecture ne $targetArchitecture) and ($targetArchitecture eq "i386")) {
  #  $pbabuilderCommand = "linux32 ".$pbabuilderCommand;
  #}
  ($main::verbosity > 0) && print "Executing this chroot command:\n".$chrootCommand."\n";
  executeOrDie($chrootCommand);

  if (length($scriptFileToRm) > 0) {
    unlink $scriptFileToRm || cleanupAndDie("FATAL: Unable to rm performed ".$workdir."/".$targetDistribution."/".$scriptFileToRm);
  }

  cleanupMounts();

  ($main::verbosity > 0) && print "Done executing things in ".$targetDistribution."\n";
}


sub buildSrpmFile( $ ) {

  my $srpmImportFile = shift;
  ($srpmImportFile eq "") && cleanupAndDie("FATAL: buildSrpmFile called w/o file parameter.\n");

  my @srpmImportDirParts = split /\//, $srpmImportFile;
  my $srpmFile = pop @srpmImportDirParts;
  my $srpmDir = join '/', @srpmImportDirParts;

  my $spoolWorkDir = "/var/spool/pba-builder/work/".$$;
  checkDirectory( $spoolWorkDir, "0755" );
  push @main::dirsToCleanup, $spoolWorkDir;
  chdir $spoolWorkDir || cleanupAndDie("FATAL: Cannot chdir to ".$spoolWorkDir."\n");

  # move .src.rpm file and childs out of brothers' way

  executeOrDie("/bin/mv ".$srpmDir."/".$srpmFile." ".$spoolWorkDir."/");

  my $buildHeadlines = "Build report for ".$srpmFile."\n\n";
  my $buildReport = "Full build report for ".$srpmFile."\n\n";

  # check srpm signature
  if (checkRpmSignature($spoolWorkDir."/".$srpmFile)) {
    $buildHeadlines .= "GOOD: Signature checked OK: ".$srpmFile."\n";
    $buildReport .= "GOOD: Signature checked OK: ".$srpmFile."\n";
  } else {
    #$buildHeadlines .= "FATAL: Signature check failed on ".$srpmFile."\n";
    #$buildReport .= "FATAL: Signature check failed on ".$srpmFile."\n";
    #reportAndDie($buildReport, $srpmMaintainers, "Signature check failed on ".$srpmFile."\n");
    $buildHeadlines .= "WARNING: Signature check failed on ".$srpmFile."\n";
    $buildReport .= "\nWARNING: Signature check failed on ".$srpmFile."\n\n";
  }

  # get spec file 
  if ($main::verbosity > 0) { $buildReport .= "Extracting spec file from ".$srpmFile."\n"; }
  executeOrDie("rpm2cpio ".$srpmFile." |cpio -i '*.spec'");
  my @specFiles = glob("*.spec");
  my $nrSpecFiles = scalar(@specFiles);
  if ($nrSpecFiles != 1) {
    cleanupAndDie("FATAL: Should have gotten 1 *.spec file but got ".$nrSpecFiles."\n");
  }
  my $specFile = $specFiles[0];

  # parse spec file 
  if ($main::verbosity > 0) { $buildReport .= "Analysing spec file ".$spoolWorkDir."/".$specFile."\n"; }
  #  my $srpmArchitecture = "";
  my $srpmCategory = "";
  my @srpmDistributions = ();
  my $srpmGeneration = "";
  my @srpmRepositories = ();
  my $srpmName = "";
  my $srpmVersion = "";
  my $srpmRelease = "";
  my $srpmFullVersion = "";
  my $srpmPackager = "";
  my $srpmMaintainers = "";
  my @srpmYumSources = ();
  my @srpmTrustedKeyURLs = ();

  open (SPECFILE, "<".$specFile) || cleanupAndDie("FATAL: Cannot read ".$specFile."\n");
  my $inHeader = 1;
  while ( my $curLine = <SPECFILE> ) {
    chomp $curLine;
    if ($inHeader) {
      if ($curLine eq "") {
        $inHeader = 0;
        next;
      }
      next if ( $curLine !~ /:/ );
      my ( $pn, $pv ) = split /: */, $curLine, 2;
      #if ($main::verbosity > 2) { $buildReport .= "pn = '".$pn."', pv='".$pv."'\n"; }
      # BuildRoot
      # Group
      # License
      if ($pn =~ /^Name$/) {
        $srpmName = $pv;
        next;
      }
      if ($pn =~ /^Packager$/) {
        $srpmPackager = $pv;
        if ($srpmMaintainers eq "") {
          $srpmMaintainers = $pv;
        } else {
          $srpmMaintainers .= ", ".$pv;
        }
        next;
      }
      if ($pn =~ /^Release$/) {
        $srpmRelease = $pv;
        $srpmFullVersion = $srpmVersion.".".$srpmRelease;
        next;
      }
      # Source
      # Summary
      if ($pn =~ /^Version$/) {
        $srpmVersion = $pv;
        $srpmFullVersion = $srpmVersion.".".$srpmRelease;
        next;
      }
      # strip ###XBCS-
      $pn =~ s/^#+X[A-Z]+-//g;
      #if ($main::verbosity > 2) { $buildReport .= "Stripped pn = '".$pn."'\n"; }
      if ($pn =~ /^PBA-Build-Source-/) {
        push @srpmYumSources, $pv;
        if ($srpmFullVersion =~ /nightly/) {
          #   http://rpm.clazzes.org/repo/pba-1.0nightly/pba-1.0nightly.repo
          # besides
          #   http://rpm.clazzes.org/repo/pba-1.0/pba-1.0.repo
          my $bsDcg = $pv;
          $bsDcg =~ s/.*\///g;
          $bsDcg =~ s/\.repo$//g;
          my $bsDcgNightly = $bsDcg."nightly";
          my $nightly = $pv;
          $nightly =~ s/$bsDcg/$bsDcgNightly/;
          if ($nightly ne $pv) {
            if ($main::verbosity > 1) { $buildReport .= "Nightly version.release detected, adding nightly build source ".$nightly."\n"; }
            push @srpmYumSources, $nightly;
          } else {
            if ($main::verbosity > 1) { 
              $buildReport .= "Nightly version.release detected but unable to nightly-ize build source ".$pv.".\n";
              $buildReport .= "  bsDcg = '".$bsDcg."'\n";
              $buildReport .= "  bsDcgNightly = '".$bsDcgNightly."'\n";
              $buildReport .= "  nightly = '".$nightly."'\n";
            }
          }
        }
        if ($srpmFullVersion =~ /beta|nightly/) {
          #   http://rpm.clazzes.org/repo/pba-1.0beta/pba-1.0beta.repo
          # besides
          #   http://rpm.clazzes.org/repo/pba-1.0/pba-1.0.repo
          my $bsDcg = $pv;
          $bsDcg =~ s/.*\///g;
          $bsDcg =~ s/\.repo$//g;
          my $bsDcgBeta = $bsDcg."beta";
          my $betaBs = $pv;
          $betaBs =~ s/$bsDcg/$bsDcgBeta/;
          if ($betaBs ne $pv) {
            if ($main::verbosity > 1) { $buildReport .= "Beta or nightly version.release detected, adding beta build source ".$betaBs."\n"; }
            push @srpmYumSources, $betaBs;
          } else {
            if ($main::verbosity > 1) { 
              $buildReport .= "Beta or nightly version.release detected but unable to beta-ize build source ".$pv.".\n";
              $buildReport .= "  bsDcg = '".$bsDcg."'\n";
              $buildReport .= "  bsDcgBeta = '".$bsDcgBeta."'\n";
              $buildReport .= "  betaBs = '".$betaBs."'\n";
            }
          }
        }
        next;
      }
      if ($pn =~ /^PBA-Category$/) {
        $srpmCategory = $pv;
        next;
      }
      if ($pn =~ /^PBA-BuildReportCc/) {
        if ($srpmMaintainers eq "") {
          $srpmMaintainers = $pv;
        } else {
          $srpmMaintainers .= ", ".$pv;
        }
        next;
      }
      if ($pn =~ /^PBA-Distributions$/) {
        @srpmDistributions = split / /, simplifyList($pv);
        next;
      }
      if ($pn =~ /^PBA-Generation$/) {
        $srpmGeneration = $pv;
        next;
      }
      if ($pn =~ /^PBA-Repository$/) {
        @srpmRepositories = split / /, simplifyList($pv);
        next;
      }
      if ($pn =~ /^PBA-TrustedKey-URL-/) {
        push @srpmTrustedKeyURLs, $pv;
        next;
      }
    } # end in header
  } # end while <SPECFILE>

  # no arch check for rpms
  ($srpmCategory eq "") && reportAndDie($buildReport, $srpmMaintainers, "No PBA-Category specified in ".$srpmFile." resp. ".$specFile."\n");
  (scalar(@srpmDistributions) < 1) && reportAndDie($buildReport, $srpmMaintainers, "No PBA-Distributions specified in ".$srpmFile." resp. ".$specFile."\n");
  ($srpmGeneration eq "") && reportAndDie($buildReport, $srpmMaintainers, "No PBA-Generation specified in ".$srpmFile." resp. ".$specFile."\n");
  foreach my $srpmDistribution (@srpmDistributions) {
    (defined $main::distributionLongNames{$srpmDistribution}) || reportAndDie($buildReport, $srpmMaintainers, "Unknown distribution ".$srpmDistribution." in ".$srpmFile." resp. ".$specFile."\n");
  }
  (scalar(@srpmRepositories) == 0) && reportAndDie($buildReport, $srpmMaintainers, "No PBA-Repositories specified in neither ".$srpmFile." resp. ".$specFile." nor for category ".$srpmCategory."\n");
  foreach my $srpmRepository (@srpmRepositories) {
    (defined $main::repositoryLongNames{$srpmRepository}) || reportAndDie($buildReport, $srpmMaintainers, "Unknown repository ".$srpmRepository." in ".$srpmFile." resp. ".$specFile."\n");
  }

  #my $buildAptRepository = $srpmRepositories[0];

  if ($main::verbosity > 1) { $buildReport .= "Package file seems ok: ".$srpmFile." resp. ".$specFile."\n"; }
  
  #my $workdir = $main::cacheDir."/".$$;

  my @resultFilesToPublish = ();
  my @logfilesToAttach = ();

  my $tarballBaseDir = $spoolWorkDir."/tarball";
  checkDirectory($tarballBaseDir);

  # iterate over distributions
  foreach my $buildDistribution (@srpmDistributions) {

    # no multiple arch support yet
    #foreach my $buildArchitecture (@srpmArchitectures) {
    my $buildArchitecture = $main::targetArchitecture; 
    
      $buildReport .= "\nBuilding for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
      chdir $spoolWorkDir || cleanupAndDie("FATAL: Cannot chdir to ".$spoolWorkDir."\n");

      # copy pbabuilder
      my $tarball = $main::libDir."/".$buildDistribution.".tar.gz";
      if (not -f $tarball) {
        $buildReport .= "\nERROR: PBuilder tarball missing: ".$tarball."\nSKIPPING build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
        $buildHeadlines .= "SKIPPED build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
        next;
      }

      my $extractPBuilderLog = $spoolWorkDir."/extract-".$buildDistribution."-".$buildArchitecture.".log";
      #push @logfilesToAttach, $extractPBuilderLog;
      extractRpmBuilder($buildDistribution, $extractPBuilderLog, $main::targetArchitecture);

      my $chrootDir = $main::cacheDir."/".$$."/".$buildDistribution;

      # prepare repos
      my $nrSrpmYumSources = scalar(@srpmYumSources);
      if ($nrSrpmYumSources > 0) {
        #checkDirectory($chrootDir."/etc/yum.repos.d");
        chdir $chrootDir."/etc/yum.repos.d" || cleanupAndDie("FATAL: Cannot chdir to ".$chrootDir."/etc/yum.repos.d/\n");
        foreach my $yumSource (@srpmYumSources) {
          my $wgetCommand = "wget --no-check-certificate ".$yumSource;
          # or curl?
          if (execute($wgetCommand)) {
            $buildReport .= "\nSuccessfully retrieved ".$yumSource." to ".pwd()."\n";
          } elsif ($srpmFullVersion =~ /beta|nightly/) {
            $buildReport .= "\nWARNING: Cannot wget ".$yumSource."\nThis is common for beta or beta builds.\n\n";
          } else {
            $buildReport .= "\nFATAL: Cannot wget ".$yumSource."\n";
            $buildHeadlines .= "FAILED to build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
            next;
          }
        }
      } else {
        $buildReport .= "\nNo yum build sources for ".$chrootDir."/etc/yum.repos.d/\n";
      }

      # inject srpm
      my $chrootDir = $main::cacheDir."/".$$."/".$buildDistribution;

      executeOrDie("mount --bind /dev ".$chrootDir."/dev");
      push @main::mountsToUnmount, $chrootDir."/dev";
      executeOrDie("mount --bind /proc ".$chrootDir."/proc");
      push @main::mountsToUnmount, $chrootDir."/proc";

      chdir $chrootDir || cleanupAndDie("FATAL: Cannot chdir to ".$chrootDir."\n");
      my $buildroot = "/usr/src/redhat";
      checkDirectory($chrootDir."/".$buildroot);
      checkDirectory($chrootDir."/".$buildroot."/BUILD");
      checkDirectory($chrootDir."/".$buildroot."/RPMS");
      checkDirectory($chrootDir."/".$buildroot."/SRPMS");
      checkDirectory($chrootDir."/".$buildroot."/SOURCES");
      checkDirectory($chrootDir."/".$buildroot."/SPECS");
      if (not execute("/bin/cp -v ".$spoolWorkDir."/".$srpmFile." ".$chrootDir."/".$buildroot."/SRPMS/")) {
        $buildReport .= "\nFATAL: Cannot copy ".$srpmDir."/".$srpmFile." to ".$chrootDir."/".$buildroot."/SRPMS/\n";
        $buildHeadlines .= "FAILED to build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
        cleanupMounts();
        next;
      }

      # log
      my $prepareLog = $spoolWorkDir."/prepare_".$buildDistribution."-".$buildArchitecture.".log";
      push @logfilesToAttach, $prepareLog;

      # inject gpg keys from TrustedKeyUrls?
      if (scalar(@srpmTrustedKeyURLs) > 0) {
        foreach my $keyURL (@srpmTrustedKeyURLs) {
          execute("wget --no-check-certificate ".$keyURL." -O ".$chrootDir."/tmp/trustedkey.asc |tee -a ".$prepareLog);
          execute("/usr/sbin/chroot ".$chrootDir." rpm --import trustedkey.asc 2>&1 |tee -a ".$prepareLog);
        }
      }

      # inject resolv.conf
      executeOrDie("/bin/cp -av /etc/resolv.conf ".$chrootDir."/etc/"." 2>&1 |tee -a ".$prepareLog);
     
      # evtl. inject proxy to yum conf
      if ($main::configParams{"HttpProxy"} ne "") {
        if (open (YUMCONF, ">>".$chrootDir."/etc/yum.conf")) {
          print YUMCONF "\n\nproxy=".$main::configParams{"HttpProxy"}."\n\n";
          close YUMCONF;
        } else {
          print "WARNING: Cannot append to ".$chrootDir."/etc/yum.conf\n";
        }
      }
      # inst srpm
      my $rpmInstSrpmCmd = "/usr/sbin/chroot ".$chrootDir.' rpm -i -v -D "_topdir '.$buildroot.'" '.$buildroot."/SRPMS/".$srpmFile." 2>&1 |tee -a ".$prepareLog;
      ($main::verbosity > 1) && print "Performing rpm install of srpm using this command: ".$rpmInstSrpmCmd."\n";
      if (not execute($rpmInstSrpmCmd)) {
        $buildReport .= "\nFATAL: rpm installation of srpm failed, see attached ".$prepareLog."\n";
        $buildHeadlines .= "FAILED to build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
        cleanupMounts();
        next;
      }

      # yum builddeps
      #my $yumBuilddepLog = $spoolWorkDir."/yum-builddep_".$buildDistribution."-".$buildArchitecture.".log";
      #push @logfilesToAttach, $yumBuilddepLog;
      my $yumBuilddebCmd = "/usr/sbin/chroot ".$chrootDir." yum-builddep -y ".$buildroot."/SRPMS/".$srpmFile. " 2>&1 |tee -a ".$prepareLog;
      ($main::verbosity > 1) && print "Performing yum builddep using this command: ".$yumBuilddebCmd."\n";
      if (not execute($yumBuilddebCmd)) {
        $buildReport .= "\nFATAL: yum-builddeps failed, see attached ".$prepareLog."\n";
        $buildHeadlines .= "FAILED to build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
        cleanupMounts();
        next;
      }

      # build
      my $rpmBuildLog = $spoolWorkDir."/rpmbuild_".$buildDistribution."-".$buildArchitecture.".log";
      push @logfilesToAttach, $rpmBuildLog;
      my $rpmBuildCmd = "/usr/sbin/chroot ".$chrootDir.' rpmbuild -bb -D "_topdir '.$buildroot.'" '.$buildroot."/SPECS/".$specFile." 2>&1 |tee -a ".$rpmBuildLog;
      # rpm signing is diffulcult, hence use signRpm() over --sign 
      ($main::verbosity > 1) && print "Performing rpmbuild using this command: ".$rpmBuildCmd."\n";
      if (not execute($rpmBuildCmd)) {
        $buildReport .= "\nFATAL: rpmbuild failed, see attached ".$rpmBuildLog."\n";
        $buildHeadlines .= "FAILED to build for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";
        cleanupMounts();
        next;
      }

      cleanupMounts();

      $buildReport .= "Build done for distribution ".$buildDistribution.", architecture ".$buildArchitecture.":\n";
      $buildHeadlines .= "SUCCESSFULLY built for distribution ".$buildDistribution.", architecture ".$buildArchitecture.".\n";

      # rename .deb files, /bin/mv them to resultdir and add them to @resultFilesToPublish, remove others
      # collect tarball
      chdir $chrootDir."/".$buildroot."/RPMS" || cleanupAndDie("FATAL: Cannot chdir to ".$chrootDir."/".$buildroot."/RPMS\n");
      #my $findCmd = "find . -type f -name '*.rpm'";
      #$buildReport .= "\nSearching *.rpm in ".pwd().":\n".`$findCmd`."\n\n";
      my @rpmFiles = ();
      my $repoArch = $buildArchitecture;
      if ($buildArchitecture eq "amd64") {
        $repoArch = "x86_64";
      }
      if (-d $repoArch) {
        chdir $chrootDir."/".$buildroot."/RPMS/".$repoArch || cleanupAndDie("FATAL: Cannot chdir to ".$chrootDir."/".$buildroot."/RPMS/".$repoArch."\n");
        my @archRpms = glob("*.rpm");
        my $nrArchRpms = scalar(@archRpms);
        if ($nrArchRpms == 0) { 
          $buildReport .= "\nWarning: No resulting binary RPMs.\n";
        } else {
          $buildReport .= "\nGot the following binary RPMs:\n";
          foreach my $binaryRpm (@archRpms) {
            $buildReport .= $binaryRpm."\n";
            push @rpmFiles, $chrootDir."/".$buildroot."/RPMS/".$repoArch."/".$binaryRpm;
          }
        }
      }
      if (-d "noarch") {
        chdir $chrootDir."/".$buildroot."/RPMS/noarch" || cleanupAndDie("FATAL: Cannot chdir to ".$chrootDir."/".$buildroot."/RPMS/noarch"."\n");
        my @noarchRpms = glob("*.rpm");
        my $nrNoarchRpms = scalar(@noarchRpms);
        if ($nrNoarchRpms == 0) { 
          $buildReport .= "\nWarning: No resulting noarch-RPMs.\n";
        } else {
          $buildReport .= "\nGot the following noarch-RPMs:\n";
          foreach my $noarchRpm (@noarchRpms) {
            $buildReport .= $noarchRpm."\n";
            push @rpmFiles, $chrootDir."/".$buildroot."/RPMS/noarch"."/".$noarchRpm;
          }
        }
      }
      my $nrResultRpms = scalar(@rpmFiles);
      #my $repodataConfFile = "";
      if ($nrResultRpms == 0) { 
        $buildReport .= "No resulting rpm to feed a repository with for distribution ".$buildDistribution."\n";
        $buildHeadlines .= "NO resulting files to feed a repository with for distribution ".$buildDistribution."\n";
      } else {
        # sign them?
        if ($main::createSignedRpms) {
          ($main::verbosity > 0) && print "Signing ".$nrResultRpms." resulting rpm files.\n";
          signRpm(join(" ", @rpmFiles));
        } else {
          ($main::verbosity > 0) && print "Not signing ".$nrResultRpms." resulting rpm files.\n";
        }
        ($main::verbosity > 0) && print "mv-ing ".$nrResultRpms." resulting rpm files.\n";
        # directory = $tarballBaseDir / category - generation / distribution / arch /
        my $cgdaDir = $tarballBaseDir."/".$srpmCategory."-".$srpmGeneration."/".$buildDistribution."/".$repoArch;
        if ($srpmFullVersion =~ /beta/) {
          $cgdaDir = $tarballBaseDir."/".$srpmCategory."-".$srpmGeneration."beta/".$buildDistribution."/".$repoArch;
        }
        if ($srpmFullVersion =~ /nightly/) {
          $cgdaDir = $tarballBaseDir."/".$srpmCategory."-".$srpmGeneration."nightly/".$buildDistribution."/".$repoArch;
        }
        checkDirectory($cgdaDir);
        if ($main::localTargetDirectory ne "") {
          executeOrDie("/bin/cp -v ".join(" ", @rpmFiles)." ".$main::localTargetDirectory."/");
        }
        executeOrDie("/bin/mv -v ".join(" ", @rpmFiles)." ".$cgdaDir."/");
      }

      execute("/bin/rm -rf ".$main::cacheDir."/".$$);

    #} # end iterate over architectures

  } # end iterate over distributions

  # check if we have RPMs at all
  chdir $tarballBaseDir || cleanupAndDie("FATAL: Cannot chdir to ".$tarballBaseDir."\n");
  my $findCmd = "find ".$tarballBaseDir." -type f -name '*.rpm'";
  open (FINDOUTPUT, $findCmd."|") || cleanupAndDie("FATAL: Cannot search. Find command was:\n".$findCmd."\npwd was: ".pwd()."\n");
  my $nrFoundRpms = 0;
  while ( my $foundRpm = <FINDOUTPUT> ) {
    $nrFoundRpms++;
  }
  if ($nrFoundRpms == 0) {
    $buildReport .= "\nERROR: No resulting RPM files at all.\n\n";
    $buildHeadlines .= "ERROR NO resulting RPM files at all.\n";
    ($main::verbosity > 0) && print $buildReport."\n".$buildHeadlines;
    report ( $buildHeadlines."\n\n".$buildReport, $srpmMaintainers, join(" ", @logfilesToAttach) );
    return;
  }

  # evtl. sign them here instead of per distro/arch?

  # repodata.conf
  my $repoDataConf = "";
  $repoDataConf .= "repositories=".join(", ", @srpmRepositories)."\n";
  my $repodataConfFile = $tarballBaseDir."/repodata.conf";
  printToFile($repodataConfFile, $repoDataConf);

  # collect tar
  my $tarLog = $spoolWorkDir."/tar_".$srpmName."_".$srpmVersion."-".$srpmRelease.".log";
  push @logfilesToAttach, $tarLog;
  my $resultTarball = $main::resultsDir."/".$srpmName."_".$srpmVersion."-".$srpmRelease.".rpms.tar.gz";
  chdir $tarballBaseDir || cleanupAndDie("FATAL: Cannot chdir to ".$tarballBaseDir."\n");
  my $tarCmd = "tar cvzf ".$resultTarball." . 2>&1 |tee ".$tarLog;
  if (not execute($tarCmd)) {
    $buildReport .= "\nFATAL: Unable to collect tarball, see attached ".$tarLog."\n";
    $buildHeadlines .= "FAILED to collect tarbarball.\n";
    ($main::verbosity > 0) && print $buildReport."\n".$buildHeadlines;
    report ( $buildHeadlines."\n\n".$buildReport, $srpmMaintainers, join(" ", @logfilesToAttach) );
    return;
  }
  $buildReport .= "\nSucessfully collected tarball ".$tarLog."\n";
  $buildHeadlines .= "SUCCESSFULLY collected tarbarball.\n";
  push @resultFilesToPublish, $resultTarball; 

  # construct $feedScripts
  my @feedScriptsToCall = ();
  my @feedScriptsToPropose = ();
  
  my @uniqFilesToPublish = uniq(@resultFilesToPublish);

  if (scalar(@uniqFilesToPublish) == 0) {
    $buildReport .= "No resulting files to feed a repository with.\n\n";
    $buildHeadlines .= "NO resulting files to feed a repository with.\n";
    ($main::verbosity > 0) && print $buildReport."\n".$buildHeadlines;
    report ( $buildHeadlines."\n\n".$buildReport, $srpmMaintainers, join(" ", @logfilesToAttach) );
    return;
  }
  
  my $uploadName = $srpmName."_".$srpmVersion."-".$srpmRelease."_rpms";

  if ($main::localTargetDirectory ne "") {

    my $feedScriptName = "feed_".$uploadName."_to_localdir.sh";
    my $feedScript = proposeMvScript( $main::localTargetDirectory, join(" ", @uniqFilesToPublish) );
    #$feedScript .= "\n/bin/rm -v \\\n";
    #foreach my $fileToPublish (@uniqFilesToPublish) {
    #  $feedScript .= "  ".$fileToPublish." \\\n";
    #}
    #$feedScript .= "  || exit 11\n";
    $feedScript .= "echo 'Done.\n'\n";
    $feedScript .= "exit 0\n";
    printToFile ($main::resultsDir."/".$feedScriptName, $feedScript, 0755);
    push @feedScriptsToCall, $main::resultsDir."/".$feedScriptName;

  } else {
  
    foreach my $srpmRepository (@srpmRepositories) {
      if ($main::repositoryFeedHostDirs{$srpmRepository} eq "") {
        $buildReport .= "I don't know how to feed this repository:\n# ".$main::repositoryLongNames{$srpmRepository}."\n\n";
        $buildHeadlines .= "SKIPPED feeding repository ".$srpmRepository."\n";
        next;
      }
      my ($feedHost, $feedDir) = split /:/, $main::repositoryFeedHostDirs{$srpmRepository}, 2;
      my $uploadDir = $feedDir."/uploading-".$uploadName."_".$$;
      my $feedScriptName = "feed_".$uploadName."_to_".$srpmRepository.".sh";
      my $feedScript = "#!/bin/sh\n";
      $feedScript = "cd ".$main::resultsDir." || exit 1\n";
      # create remote upload dir
      $feedScript .= "echo 'Calling ...\n/usr/bin/ssh ".$feedHost." /bin/mkdir -pv ".$uploadDir."'\n";
      $feedScript .= "/usr/bin/ssh ".$feedHost." /bin/mkdir -pv ".$uploadDir." \\\n";
      $feedScript .= "  || /usr/bin/ssh ".$feedHost." /bin/mkdir -pv ".$uploadDir." \\\n";
      $feedScript .= "  || exit 2\n";
      $feedScript .= "echo 'Done.\n'\n";
      # echo /usr/bin/scp files
      $feedScript .= "echo 'Calling ...\n/usr/bin/scp \\\n";
      foreach my $fileToPublish (@uniqFilesToPublish) {
        $feedScript .= "  ".$fileToPublish." \\\n";
      }
      $feedScript .= "  ".$feedHost.":".$uploadDir."/'\n";
      # /usr/bin/scp files
      $feedScript .= "/usr/bin/scp \\\n";
      foreach my $fileToPublish (@uniqFilesToPublish) {
        $feedScript .= "  ".$fileToPublish." \\\n";
      }
      $feedScript .= "  ".$feedHost.":".$uploadDir."/ \\\n";
      $feedScript .= "  || /usr/bin/scp \\\n";
      foreach my $fileToPublish (@uniqFilesToPublish) {
        $feedScript .= "  ".$fileToPublish." \\\n";
      }
      $feedScript .= "  ".$feedHost.":".$uploadDir."/ \\\n";
      $feedScript .= "  || exit 3\n";
      $feedScript .= "echo 'Done.\n'\n";
      # /bin/mv files
      $feedScript .= "echo 'Calling ...\n/usr/bin/ssh ".$feedHost." /bin/mv -v ".$uploadDir."/* ".$feedDir."/'\n";
      $feedScript .= "/usr/bin/ssh ".$feedHost." /bin/mv -v ".$uploadDir."/* ".$feedDir."/ \\\n";
      $feedScript .= "  || /usr/bin/ssh ".$feedHost." /bin/mv -v ".$uploadDir."/* ".$feedDir."/ \\\n";
      $feedScript .= "  || exit 4\n";
      $feedScript .= "echo 'Done.\n'\n";
      # /bin/rm remote uploading dir
      $feedScript .= "echo 'Calling ...\n/usr/bin/ssh ".$feedHost." /bin/rmdir -v ".$uploadDir."'\n";
      $feedScript .= "/usr/bin/ssh ".$feedHost." /bin/rmdir -v ".$uploadDir." \\\n";
      $feedScript .= "  || /usr/bin/ssh ".$feedHost." /bin/rmdir -v ".$uploadDir." \\\n";
      $feedScript .= "  || exit 5\n";
      $feedScript .= "echo 'Done.\n'\n";
      # /bin/rm files if we are the last feed script
      $feedScript .= "# if we are the only feed script left, remove files\n";
      $feedScript .= 'OTHERFEEDERS=$(ls -1 |egrep "feed_'.$uploadName.'.*\.sh$" |egrep -v '.$feedScriptName.")\n";
      $feedScript .= 'if [ -z "$OTHERFEEDERS" ] ; then'."\n";
      # echo /bin/rm
      $feedScript .= "  echo 'Calling ...\n/bin/rm -v \\\n";
      foreach my $fileToPublish (@uniqFilesToPublish) {
        $feedScript .= "    ".$fileToPublish." \\\n";
      }
      $feedScript .= "'\n";
      # rm
      $feedScript .= "  /bin/rm -v \\\n";
      foreach my $fileToPublish (@uniqFilesToPublish) {
        $feedScript .= "    ".$fileToPublish." \\\n";
      }
      $feedScript .= "    || exit 6\n";
      $feedScript .= "  echo 'Done.\n'\n";
      $feedScript .= "else\n";
      $feedScript .= "  echo 'Not removing ...\n";
      foreach my $fileToPublish (@uniqFilesToPublish) {
        $feedScript .= "    ".$fileToPublish."\n";
      }
      $feedScript .= "... because I found the following other feeders:'\n";
      $feedScript .= '  echo "$OTHERFEEDERS"'."\n";
      $feedScript .= "fi\n";
      #$feedScript .= "# finally, suicide\n";
      #$feedScript .= "/bin/rm -fv ".$feedScriptName."\n";
      $feedScript .= "exit 0\n";
      if ($main::repositoryAutoFeed{$srpmRepository} eq "true") {
        push @feedScriptsToCall, $main::resultsDir."/".$feedScriptName;
      } else {
        push @feedScriptsToPropose, $main::resultsDir."/".$feedScriptName;
      }
      printToFile ($main::resultsDir."/".$feedScriptName, $feedScript, 0755);
    }
  }

  if (scalar(@feedScriptsToCall) > 0) {
    foreach my $feedScript (@feedScriptsToCall) {
      my $feedScriptBase = $feedScript;
      $feedScriptBase =~ s/^.*\///g;
      my $feedLog = $spoolWorkDir."/".$feedScriptBase.".log";
      push @logfilesToAttach, $feedLog;
      my $feedRc = system("/bin/sh ".$feedScript." >".$feedLog." 2>&1");
      if ($feedRc == 0) {
        $buildReport .= "Successfully executed this feed script:\n".$feedScript."\n";
        $buildHeadlines .= "SUCCESSFULLY executed ".$feedScript."\n";
        if (-f $feedScript) { unlink $feedScript; }
      } else {
        $buildReport .= "ERROR: Failed executing this feed script (rc=".$feedRc."):\n".$feedScript."\n";
        $buildHeadlines .= "FAILED to execute ".$feedScript."\n";
      }
    }
  }

  if (scalar(@feedScriptsToPropose) > 0) {
    $buildReport .= "\nProposing to call these feed-and-cleanup scripts:\n";
    $buildHeadlines .= "\nAuto-Feed off, to publish PLEASE CALL these feed-and-cleanup scripts:\n";
    foreach my $feedScript (@feedScriptsToPropose) {
      $buildReport .= $feedScript." && \\\n  /bin/rm -fv ".$feedScript."\n";
      $buildHeadlines .= $feedScript." && \\\n  /bin/rm -fv ".$feedScript."\n";
    }
  }
 
  ($main::verbosity > 0) && print $buildReport."\n".$buildHeadlines;
  report ( $buildHeadlines."\n\n".$buildReport, $srpmMaintainers, join(" ", @logfilesToAttach) );

  #if ($main::verbosity <= 2) {
 
    unlink @logfilesToAttach;
 
    # /bin/rm main srpm, spec (in any case!)
    chdir $spoolWorkDir || cleanupAndDie("FATAL: Cannot chdir to ".$spoolWorkDir."\n");
    unlink $srpmFile || cleanupAndDie("FATAL: Cannot /bin/rm ".$spoolWorkDir."/".$srpmFile."\n");
    unlink $specFile || cleanupAndDie("FATAL: Cannot /bin/rm ".$spoolWorkDir."/".$specFile."\n");
    chdir "..";
    executeOrDie("/bin/rm -rfv ".$spoolWorkDir);

  #}

  ($main::verbosity > 1) && print "Done building ".$srpmFile."\n";

}


1;

